/* AlignmentViewer.java
 *
 * created: Mon Jul 12 1999
 *
 * This file is part of Artemis
 *
 * Copyright (C) 1999,2000,2001  Genome Research Limited
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/pathsoft/repository/powmap/diana/components/AlignmentViewer.java,v 1.31 2001/09/14 14:43:53 kmr Exp $
 */

package diana.components;

import diana.*;
import diana.sequence.*;

import uk.ac.sanger.pathogens.OutOfRangeException;
import uk.ac.sanger.pathogens.embl.Range;

import java.awt.*;
import java.awt.event.*;

/**
 *  This component shows an alignment of two sequences using the data from a
 *  ComparisonData object.
 *
 *  @author Kim Rutherford
 *  @version $Id: AlignmentViewer.java,v 1.31 2001/09/14 14:43:53 kmr Exp $
 **/

public class AlignmentViewer extends CanvasPanel
    implements SequenceChangeListener {
  /**
   *  Create a new AlignmentViewer for the given entries.
   *  @param subject_feature_display The FeatureDisplay that is above this
   *    component.
   *  @param query_feature_display The FeatureDisplay that is below this
   *    component.
   *  @param comparison_data Provides the AlignMatch objects that will be
   *    displayed.
   **/
  public AlignmentViewer (final FeatureDisplay subject_feature_display,
                          final FeatureDisplay query_feature_display,
                          final ComparisonData comparison_data) {
    this.subject_feature_display = subject_feature_display;
    this.query_feature_display   = query_feature_display;
    this.comparison_data         = comparison_data;
    this.all_matches             = comparison_data.getMatches ();

    subject_entry_group          = getSubjectDisplay ().getEntryGroup ();
    query_entry_group            = getQueryDisplay ().getEntryGroup ();

    final Bases subject_bases = getSubjectForwardStrand ().getBases ();
    final Bases query_bases   = getQueryForwardStrand ().getBases ();

    final Selection subject_selection = getSubjectDisplay ().getSelection ();
    final Selection query_selection = getQueryDisplay ().getSelection ();

    final SelectionChangeListener subject_listener =
      new SelectionChangeListener () {
        public void selectionChanged (SelectionChangeEvent event) {
          final Range range = subject_selection.getSelectionRange ();
          selectFromSubjectRange (range);
        }
      };

    final SelectionChangeListener query_listener =
      new SelectionChangeListener () {
        public void selectionChanged (SelectionChangeEvent event) {
          final Range range = query_selection.getSelectionRange ();
          selectFromQueryRange (range);
        }
      };

    subject_selection.addSelectionChangeListener (subject_listener);
    query_selection.addSelectionChangeListener (query_listener);

    subject_bases.addSequenceChangeListener (this, 0);
    query_bases.addSequenceChangeListener (this, 0);

    orig_subject_forward_strand = getSubjectForwardStrand ();
    orig_query_forward_strand = getQueryForwardStrand ();

    score_colours = makeColours ();

    getCanvas ().addMouseListener (new MouseAdapter () {
      public void mousePressed (final MouseEvent event) {
        handleCanvasMousePress (event);
      }
    });

    getCanvas ().addMouseMotionListener (new MouseMotionAdapter () {
      public void mouseDragged (final MouseEvent event) {
        if (!modifiersForLockToggle (event)) {
          setSelection (event.getPoint ());
          repaintCanvas ();
        }
      }
    });

    scroll_bar = new Scrollbar (Scrollbar.VERTICAL, 1, 1, 1, 1000);
    scroll_bar.setBlockIncrement (10);

    scroll_bar.addAdjustmentListener (new AdjustmentListener () {
      public void adjustmentValueChanged(AdjustmentEvent e) {
        repaintCanvas ();
      }
    });

    add (scroll_bar, "East");
  }

  /**
   *  Returns true if and only if the given MouseEvent should toggle the lock
   *  displays toggle.
   **/
  private boolean modifiersForLockToggle (final MouseEvent event) {
    return (event.getModifiers () & InputEvent.BUTTON2_MASK) != 0 ||
      event.isAltDown () || event.isControlDown ();
  }

  /**
   *  Select those matches that overlap the given range on the subject
   *  sequence.
   **/
  public void selectFromSubjectRange (final Range select_range) {
    if (select_range == null) {
      return;
    }

    if (disable_selection_from_ranges) {
      return;
    }

    selected_matches = null;

    for (int i = 0 ; i < all_matches.length ; ++i) {
      final AlignMatch this_match = all_matches [i];

      if (!checkMatch (this_match)) {
        // not visible
        continue;
      }

      int subject_sequence_start =
        this_match.getRealSubjectSequenceStart (getOrigSubjectForwardStrand ());
      int subject_sequence_end =
        this_match.getRealSubjectSequenceEnd (getOrigSubjectForwardStrand ());

      if (subject_sequence_end < subject_sequence_start) {
        final int tmp = subject_sequence_start;
        subject_sequence_start = subject_sequence_end;
        subject_sequence_end = tmp;
      }

      if (select_range.getStart () < subject_sequence_start &&
          select_range.getEnd () < subject_sequence_start) {
        continue;
      }

      if (select_range.getStart () > subject_sequence_end &&
          select_range.getEnd () > subject_sequence_end) {
        continue;
      }

      if (selected_matches == null) {
        selected_matches = new AlignMatchVector ();
      }

      selected_matches.add (this_match);
    }

    selectionChanged ();
  }

  /**
   *  Select those matches that overlap the given range on the query sequence.
   **/
  public void selectFromQueryRange (final Range select_range) {
    if (select_range == null) {
      return;
    }

    if (disable_selection_from_ranges) {
      return;
    }

    selected_matches = null;

    for (int i = 0 ; i < all_matches.length ; ++i) {
      final AlignMatch this_match = all_matches [i];

      if (!checkMatch (this_match)) {
        // not visible
        continue;
      }

      int query_sequence_start =
        this_match.getRealQuerySequenceStart (getOrigQueryForwardStrand ());
      int query_sequence_end =
        this_match.getRealQuerySequenceEnd (getOrigQueryForwardStrand ());

      if (query_sequence_end < query_sequence_start) {
        final int tmp = query_sequence_start;
        query_sequence_start = query_sequence_end;
        query_sequence_end = tmp;
      }

      if (select_range.getStart () < query_sequence_start &&
          select_range.getEnd () < query_sequence_start) {
        continue;
      }

      if (select_range.getStart () > query_sequence_end &&
          select_range.getEnd () > query_sequence_end) {
        continue;
      }

      if (selected_matches == null) {
        selected_matches = new AlignMatchVector ();
      }

      selected_matches.add (this_match);
    }

    selectionChanged ();
  }

  /**
   *  Select the given match and move it to the top of the display.
   **/
  public void setSelection (final AlignMatch match) {
    selected_matches = new AlignMatchVector ();

    selected_matches.add (match);

    selectionChanged ();
  }

  /**
   *  This method tells this AlignmentViewer component where the subject
   *  sequence is now.
   **/
  public void setSubjectSequencePosition (final DisplayAdjustmentEvent event) {
    last_subject_event = event;
    repaintCanvas ();
  }

  /**
   *  This method tells this AlignmentViewer component where the query
   *  sequence is now.
   **/
  public void setQuerySequencePosition (final DisplayAdjustmentEvent event) {
    last_query_event = event;
    repaintCanvas ();
  }

  /**
   *  Implementation of the SequenceChangeListener interface.  The display is
   *  redrawn if there is an event.
   **/
  public void sequenceChanged (final SequenceChangeEvent event) {
    repaintCanvas ();
  }

  /**
   *  Handle a mouse press event on the drawing canvas - select on click,
   *  select and broadcast it on double click.
   **/
  private void handleCanvasMousePress (final MouseEvent event) {
    if (event.isPopupTrigger () || event.isMetaDown ()) {
      final PopupMenu popup = new PopupMenu ();

      final MenuItem alignmatch_list_item =
        new MenuItem ("View Selected Matches");

      popup.add (alignmatch_list_item);

      alignmatch_list_item.addActionListener (new ActionListener () {
        public void actionPerformed (ActionEvent _) {
          if (selected_matches == null) {
            new MessageFrame ("No matches selected").setVisible (true);
          } else {
            final AlignMatchVector matches =
              (AlignMatchVector) selected_matches.clone ();

            final AlignMatchViewer viewer =
              new AlignMatchViewer (AlignmentViewer.this, matches);

            viewer.setVisible (true);
          }
        }
      });

      final MenuItem flip_query_item =
        new MenuItem ("RevComp Query Sequence");

      popup.add (flip_query_item);

      flip_query_item.addActionListener (new ActionListener () {
        public void actionPerformed (ActionEvent _) {
          try {
            getQueryEntryGroup ().reverseComplement ();
            if (selected_matches != null && selected_matches.size () > 0) {
              // pick one so the user doesn't get completely lost
              alignAt (selected_matches.elementAt (0));
            }
          } catch (uk.ac.sanger.pathogens.ReadOnlyException e) {
            new MessageFrame ("Query sequence is read-only").setVisible (true);
          }
        }
      });

      final MenuItem lock_item = new MenuItem ("Lock Sequences");

      popup.add (lock_item);

      lock_item.addActionListener (new ActionListener () {
        public void actionPerformed (ActionEvent _) {
          lockDisplays ();
        }
      });

      final MenuItem unlock_item = new MenuItem ("Unlock Sequences");

      popup.add (unlock_item);

      unlock_item.addActionListener (new ActionListener () {
        public void actionPerformed (ActionEvent _) {
          unlockDisplays ();
        }
      });

      final MenuItem cutoffs_item = new MenuItem ("Set Cutoffs ...");

      popup.add (cutoffs_item);

      cutoffs_item.addActionListener (new ActionListener () {
        public void actionPerformed (ActionEvent _) {
          final ScoreChangeListener minimum_listener =
            new ScoreChangeListener () {
              public void scoreChanged (final ScoreChangeEvent event) {
                minimum_score = event.getValue ();
                repaintCanvas ();
              }
            };

          final ScoreChangeListener maximum_listener =
            new ScoreChangeListener () {
              public void scoreChanged (final ScoreChangeEvent event) {
                maximum_score = event.getValue ();
                repaintCanvas ();
              }
            };

          final ScoreChanger score_changer =
            new ScoreChanger ("Score Cutoffs",
                              minimum_listener, maximum_listener);

          score_changer.setVisible (true);
        }
      });

      popup.addSeparator ();

      final CheckboxMenuItem offer_to_flip_item =
        new CheckboxMenuItem ("Offer To RevComp", offer_to_flip_flag);

      offer_to_flip_item.addItemListener (new ItemListener () {
        public void itemStateChanged(ItemEvent e) {
          offer_to_flip_flag = !offer_to_flip_flag;
        }
      });

      popup.add (offer_to_flip_item);

      getCanvas ().add (popup);
      popup.show (getCanvas (), event.getX (), event.getY ());
    } else {
      if (event.getID() != MouseEvent.MOUSE_PRESSED) {
        return;
      }

      if (event.getClickCount () == 2) {
        handleCanvasDoubleClick (event);
      } else {
        handleCanvasSingleClick (event);
      }
    }

    repaintCanvas ();
  }

  /**
   *  Handle a double click on the canvas.
   **/
  private void handleCanvasDoubleClick (final MouseEvent event) {
    if (selected_matches != null) {
      // there should be only one match in the array because
      // handleCanvasSingleClick () has been called
      alignAt (selected_matches.elementAt (0));
    }
  }

  /**
   *  Send an AlignmentEvent to all the AlignmentListeners.
   *  @param align_match The AlignMatch that we have just centred on.
   **/
  public void alignAt (final AlignMatch align_match) {
    final java.util.Vector targets;
    // copied from a book - synchronizing the whole method might cause a
    // deadlock
    synchronized (this) {
      targets = (java.util.Vector) alignment_event_listeners.clone ();
    }

    for (int i = 0 ; i < targets.size () ; ++i) {
      final AlignmentListener listener =
        (AlignmentListener) targets.elementAt (i);

      listener.alignMatchChosen (new AlignmentEvent (align_match));
    }
  }

  /**
   *  Handle a single click on the canvas.
   **/
  private void handleCanvasSingleClick (final MouseEvent event) {
    if (modifiersForLockToggle (event)) {
      toggleDisplayLock ();
    } else {
      setSelection (event.getPoint ());
    }
  }

  /**
   *  Set the selection to be the match at the given point.
   **/
  private void setSelection (final Point point) {
    if (last_subject_event == null || last_query_event == null) {
      return;
    }

    final AlignMatch clicked_align_match =
      getAlignMatchFromPosition (point);

    if (clicked_align_match == null) {
      selected_matches = null;
    } else {
      selected_matches = new AlignMatchVector ();

      selected_matches.add (clicked_align_match);
    }

    selectionChanged ();
  }

  /**
   *  Return the AlignMatch at the given Point on screen or null if there is
   *  no match at that point.  The alignment_data_array is searched in reverse
   *  order.
   **/
  private AlignMatch getAlignMatchFromPosition (final Point click_point) {
    final int canvas_height = getCanvas ().getSize ().height;
    final int canvas_width = getCanvas ().getSize ().width;

    for (int i = all_matches.length - 1 ; i >= 0  ; --i) {
      final AlignMatch this_match = all_matches [i];

      final int [] match_x_positions =
        getMatchCoords (canvas_width, this_match);

      if (match_x_positions == null) {
        continue;
      }

      if (!checkMatch (this_match)) {
        continue;
      }

      final int subject_start_x = match_x_positions[0];
      final int subject_end_x = match_x_positions[1];
      final int query_start_x = match_x_positions[2];
      final int query_end_x = match_x_positions[3];

      // this is the x coordinate of the point where the line y = click_point
      // hits the left edge of the match box
      final double match_left_x =
        subject_start_x +
        (1.0 * (query_start_x - subject_start_x)) *
        (1.0 * click_point.y / canvas_height);

      // this is the x coordinate of the point where the line y = click_point
      // hits the right edge of the match box
      final double match_right_x =
        subject_end_x +
        (1.0 * (query_end_x - subject_end_x)) *
        (1.0 * click_point.y / canvas_height);

      if (click_point.x >= match_left_x - 1 &&
          click_point.x <= match_right_x + 1 ||
          click_point.x <= match_left_x + 1 &&
          click_point.x >= match_right_x - 1) {
        return this_match;
      }
    }

    return null;
  }

  /**
   *  This method is called by setSelection () and others whenever the list of
   *  selected/highlighted hits changes.  Calls alignmentSelectionChanged ()
   *  on all interested AlignmentSelectionChangeListener objects, moves the
   *  selected matches to the top of the display and then calls
   *  repaintCanvas ().
   **/
  private void selectionChanged () {
    for (int i = 0 ; i < selection_change_listeners.size () ; ++i) {
      final AlignMatchVector matches =
        (AlignMatchVector) selected_matches.clone ();

      final AlignmentSelectionChangeEvent ev =
        new AlignmentSelectionChangeEvent (this, matches);

      final AlignmentSelectionChangeListener listener =
        (AlignmentSelectionChangeListener) selection_change_listeners.elementAt (i);

      listener.alignmentSelectionChanged (ev);
    }

    if (selected_matches != null && selected_matches.size () > 0) {
      // a count of the number of selected matches seen so far
      int seen_and_selected_count = 0;

      for (int i = 0 ; i < all_matches.length ; ++i) {
        final AlignMatch this_match = all_matches[i];

        if (selected_matches.contains (this_match)) {
          ++seen_and_selected_count;
        } else {
          if (seen_and_selected_count > 0) {
            // move the matches down to fill the gap
            all_matches[i-seen_and_selected_count] = all_matches[i];
          }
        }
      }

      // put the selected_matches at the end of all_matches
      for (int i = 0 ; i < selected_matches.size () ; ++i) {
        all_matches[all_matches.length - selected_matches.size () + i] =
          selected_matches.elementAt (i);
      }
    }

    repaintCanvas ();
  }

  /**
   *  Add the AlignmentSelectionChangeListener to the list of objects
   *  listening for AlignmentSelectionChangeEvents.
   **/
  public void
    addAlignmentSelectionChangeListener (final AlignmentSelectionChangeListener l) {
    selection_change_listeners.addElement (l);
  }

  /**
   *  Remove the AlignmentSelectionChangeListener from the list of objects
   *  listening for AlignmentSelectionChangeEvents.
   **/
  public void
    removeAlignmentSelectionChangeListener (final AlignmentSelectionChangeListener l) {
    selection_change_listeners.removeElement (l);
  }

  /**
   *  Adds the specified AlignmentEvent listener to receive events from this
   *  object.
   *  @param l the listener.
   **/
  public void addAlignmentListener (AlignmentListener l) {
    alignment_event_listeners.addElement (l);
  }

  /**
   *  Removes the specified AlignmentEvent listener so that it no longer
   *  receives events from this object.
   *  @param l the listener.
   **/
  public void removeAlignmentListener (AlignmentListener l) {
    alignment_event_listeners.removeElement (l);
  }

  /**
   *  Returns true if and only if we should offer to flip the query
   *  sequence when the user double clicks on a flipped match.
   **/
  public boolean offerToFlip () {
    return offer_to_flip_flag;
  }

  /**
   *  Call repaint () on the canvas.
   **/
  private void repaintCanvas () {
    getCanvas ().repaint ();
  }

  /**
   *  The main paint function for the canvas.  An off screen image used for
   *  double buffering when drawing the canvas.
   *  @param g The Graphics object of the canvas.
   **/
  protected void paintCanvas (final Graphics g) {
    fillBackground (g);

    if (last_subject_event != null && last_query_event != null) {
      drawAlignments (g);
      drawLabels (g);
    }
  }

  /**
   *  Draw the background colour of the frames.
   **/
  private void fillBackground (final Graphics g) {
    g.setColor (Color.white);

    g.fillRect (0, 0, getCanvas ().getSize ().width,
                getCanvas ().getSize ().height);
  }

  /**
   *  Draw the labels into the given Graphics object.  There is a label at the
   *  top left showing info about the current AlignMatch object,
   *
   *  XXX
   *  a label at
   *  the bottom left showing whether or not the query sequence is reverse
   *  complemented
   *  XXX
   *
   *  and a label beside the Scrollbar showing the current score
   *  cutoff.
   **/
  private void drawLabels (final Graphics g) {
    final FontMetrics fm = g.getFontMetrics ();
    final int canvas_width = getCanvas ().getSize ().width;
    final int canvas_height = getCanvas ().getSize ().height;

    final String cutoff_label =
      Integer.toString (scroll_bar.getValue ());

    final int cutoff_label_width = fm.stringWidth (cutoff_label);

    int cutoff_label_position =
      (int)((scroll_bar.getValue () -
             scroll_bar.getMinimum ()) / (1.0 *
            (scroll_bar.getMaximum () -
             scroll_bar.getMinimum ())) * canvas_height);

    if (cutoff_label_position < getFontAscent ()) {
      cutoff_label_position = getFontAscent ();
    }


    final int [] cutoff_x_points = {
      canvas_width - cutoff_label_width,
      canvas_width - 2,
      canvas_width - 2,
      canvas_width - cutoff_label_width,
    };

    final int [] cutoff_y_points = {
      cutoff_label_position + 1,
      cutoff_label_position + 1,
      cutoff_label_position - getFontAscent (),
      cutoff_label_position - getFontAscent (),
    };

    g.setColor (Color.white);
    g.fillPolygon (cutoff_x_points, cutoff_y_points, 4);

    g.setColor (Color.black);

    g.drawString (cutoff_label, canvas_width - cutoff_label_width,
                  cutoff_label_position);


    if (selected_matches != null) {
      final String match_string;

      if (selected_matches.size () > 1) {
        match_string = selected_matches.size () + " matches selected";
      } else {
        final AlignMatch selected_align_match = selected_matches.elementAt (0);

        match_string =
          selected_align_match.getQuerySequenceStart () + ".." +
          selected_align_match.getQuerySequenceEnd () + " -> " +
          selected_align_match.getSubjectSequenceStart () + ".." +
          selected_align_match.getSubjectSequenceEnd () + " score " +
          selected_align_match.getScore ();
      }

      final int match_string_width = fm.stringWidth (match_string);

      final int [] match_x_points = {
        0, 0, match_string_width, match_string_width
      };

      final int [] match_y_points = {
        0, getFontAscent () + 1, getFontAscent () + 1, 0
      };

      g.setColor (Color.white);
      g.fillPolygon (match_x_points, match_y_points, 4);

      g.setColor (Color.black);

      g.drawString (match_string, 0, getFontAscent ());
    }

    String status_string;

    if (displaysAreLocked ()) {
      status_string = "LOCKED";
    } else {
      status_string = null;
    }

    final String revcomp_string;

    if (queryIsRevComp ()) {
      if (subjectIsRevComp ()) {
        revcomp_string = "Query and Subject: RevComp";
      } else {
        revcomp_string = "Query: RevComp";
      }
    } else {
      if (subjectIsRevComp ()) {
        revcomp_string = "Subject: RevComp";
      } else {
        revcomp_string = null;
      }
    }

    if (revcomp_string != null) {
      if (status_string == null) {
        status_string = revcomp_string;
      } else {
        status_string += "  -  " + revcomp_string;
      }
    }

    if (status_string != null) {
      final int status_string_width = fm.stringWidth (status_string);

      final int [] revcomp_x_points = {
        0, 0, status_string_width, status_string_width
      };

      final int [] revcomp_y_points = {
        canvas_height, canvas_height - getFontAscent () - 1,
        canvas_height - getFontAscent () - 1, canvas_height
      };

      g.setColor (Color.white);
      g.fillPolygon (revcomp_x_points, revcomp_y_points, 4);

      g.setColor (Color.black);

      g.drawString (status_string, 0, canvas_height - 2);
    }
  }

  /**
   *  Draw the alignments into the given Graphics object.
   **/
  private void drawAlignments (final Graphics g) {
    final int canvas_height = getCanvas ().getSize ().height;
    final int canvas_width = getCanvas ().getSize ().width;

    final Range subject_visible_range =
      getSubjectDisplay ().getVisibleRange ();
    final Range query_visible_range =
      getQueryDisplay ().getVisibleRange ();

    for (int i = 0 ; i < all_matches.length ; ++i) {
      final AlignMatch this_match = all_matches [i];

      final int [] match_x_positions =
        getMatchCoords (canvas_width, this_match);

      if (match_x_positions == null) {
        continue;
      }

      final int subject_start_x = match_x_positions[0];
      final int subject_end_x = match_x_positions[1];
      final int query_start_x = match_x_positions[2];
      final int query_end_x = match_x_positions[3];

      final int [] x_coords = new int [4];
      final int [] y_coords = new int [4];

      x_coords[0] = subject_start_x;
      y_coords[0] = 0;
      x_coords[1] = query_start_x;
      y_coords[1] = canvas_height;
      x_coords[2] = query_end_x;
      y_coords[2] = canvas_height;
      x_coords[3] = subject_end_x;
      y_coords[3] = 0;

      if (!checkMatch (this_match)) {
        // not visible
        continue;
      }

      final boolean highlight_this_match;

      if (selected_matches != null &&
          selected_matches.contains (this_match)) {
        highlight_this_match = true;
      } else {
        highlight_this_match = false;
      }

      final int OFFSCREEN = 3000;

      final int score = this_match.getScore ();

      if (highlight_this_match) {
        g.setColor (Color.yellow);
      } else {
        if (score == -1) {
          g.setColor (Color.red);
        } else {
          int colour_index = score_colours.length - 1;

          if (maximum_score > minimum_score) {
            colour_index =
              (int)(score_colours.length * 0.999 *
                    (score - minimum_score) /
                    (maximum_score - minimum_score));
          }

          g.setColor (score_colours[colour_index]);
        }
      }

      g.fillPolygon (x_coords, y_coords, x_coords.length);

      if (subject_end_x - subject_start_x < 5 &&
          subject_end_x - subject_start_x > -5 ||
          subject_start_x < -OFFSCREEN ||
          subject_end_x > OFFSCREEN ||
          query_start_x < -OFFSCREEN ||
          query_end_x > OFFSCREEN) {

        // match is (probably) narrow

        // XXX
      } else {

        // draw a black outline the make the match stand out
        g.setColor (Color.black);
        g.drawLine (subject_start_x, 0, query_start_x, canvas_height);
        g.drawLine (subject_end_x, 0, query_end_x, canvas_height);
      }

    }
  }

  /**
   *  Return true if and only if the given match is currently visible.
   **/
  private boolean checkMatch (final AlignMatch match) {
    final int score = match.getScore ();

    if (score > -1) {
      if (score < minimum_score || score > maximum_score) {
        return false;
      }
    }

    if (Math.abs (match.getSubjectSequenceStart () -
                  match.getSubjectSequenceEnd ()) <
        scroll_bar.getValue ()) {
      return false;
    }

    return true;
  }

  /**
   *  Return the screen x positions of the corners of the given match.  The
   *  order is Top Left, Top Right, Bottom Left, Bottom Right, unless the
   *  match is an inversion, in which case it will be TL,TR,BR,BL.  Returns
   *  null if and only if the match is not currently visible.
   **/
  private int [] getMatchCoords (final int canvas_width,
                                 final AlignMatch this_match) {
    int subject_sequence_start =
      this_match.getRealSubjectSequenceStart (getOrigSubjectForwardStrand ());
    int subject_sequence_end =
      this_match.getRealSubjectSequenceEnd (getOrigSubjectForwardStrand ());
    int query_sequence_start =
      this_match.getRealQuerySequenceStart (getOrigQueryForwardStrand ());
    int query_sequence_end =
      this_match.getRealQuerySequenceEnd (getOrigQueryForwardStrand ());

    // add one base because we want to draw to the end of the base
    if (subjectIsRevComp ()) {
      subject_sequence_start += 1;
    } else {
      subject_sequence_end += 1;
    }

    if (this_match.isRevMatch () && !queryIsRevComp () ||
        !this_match.isRevMatch () && queryIsRevComp ()) {
      query_sequence_start += 1;
    } else {
      query_sequence_end += 1;
    }

    final int subject_start_x =
      getScreenPosition (canvas_width, last_subject_event,
                         subject_sequence_start);
    final int subject_end_x =
      getScreenPosition (canvas_width, last_subject_event,
                         subject_sequence_end);

    final int query_start_x =
      getScreenPosition (canvas_width, last_query_event,
                         query_sequence_start);
    final int query_end_x =
      getScreenPosition (canvas_width, last_query_event,
                         query_sequence_end);

    boolean subject_off_left = false;
    boolean subject_off_right = false;
    boolean query_off_left = false;
    boolean query_off_right = false;

    if (subject_start_x < 0 && subject_end_x < 0) {
      subject_off_left = true;
    }

    if (subject_start_x >= canvas_width && subject_end_x >= canvas_width) {
      subject_off_right = true;
    }

    if (query_start_x < 0 && query_end_x < 0) {
      query_off_left = true;
    }

    if (query_start_x >= canvas_width && query_end_x >= canvas_width) {
      query_off_right = true;
    }

    if ((subject_off_left ? 1 : 0) +
        (query_off_left ? 1 : 0) +
        (subject_off_right ? 1 : 0) +
        (query_off_right ? 1 : 0) == 2) {
      return null;
    } else {
      final int [] return_values = new int [4];

      return_values[0] = subject_start_x;
      return_values[1] = subject_end_x;
      return_values[2] = query_start_x;
      return_values[3] = query_end_x;

      return return_values;
    }
  }

  /**
   *  Return the current forward Strand of the subject EntryGroup.
   **/
  private Strand getSubjectForwardStrand () {
    return getSubjectEntryGroup ().getBases ().getForwardStrand ();
  }

  /**
   *  Return the current forward Strand of the query EntryGroup.
   **/
  private Strand getQueryForwardStrand () {
    return getQueryEntryGroup ().getBases ().getForwardStrand ();
  }

  /**
   *  Return the current reverse Strand of the subject EntryGroup.
   **/
  private Strand getSubjectReverseStrand () {
    return getSubjectEntryGroup ().getBases ().getReverseStrand ();
  }

  /**
   *  Return the current reverse Strand of the query EntryGroup.
   **/
  private Strand getQueryReverseStrand () {
    return getQueryEntryGroup ().getBases ().getReverseStrand ();
  }

  /**
   *  Return the subject EntryGroup that was passed to the constructor.
   **/
  private EntryGroup getSubjectEntryGroup () {
    return subject_entry_group;
  }

  /**
   *  Return the subject EntryGroup that was passed to the constructor.
   **/
  private EntryGroup getQueryEntryGroup () {
    return query_entry_group;
  }

  /**
   *  Return the reference of the subject FeatureDisplay.
   **/
  public FeatureDisplay getSubjectDisplay () {
    return subject_feature_display;
  }

  /**
   *  Return the reference of the query FeatureDisplay.
   **/
  public FeatureDisplay getQueryDisplay () {
    return query_feature_display;
  }

  /**
   *  Returns true if and only if the subject sequence has been flipped since
   *  this object was created.
   **/
  private boolean subjectIsRevComp () {
    final Strand current_subject_forward_strand = getSubjectForwardStrand ();

    if (getOrigSubjectForwardStrand () == current_subject_forward_strand) {
      return false;
    } else {
      return true;
    }
  }

  /**
   *  Returns true if and only if the query sequence has been flipped since
   *  this object was created.
   **/
  private boolean queryIsRevComp () {
    final Strand current_query_forward_strand = getQueryForwardStrand ();

    if (getOrigQueryForwardStrand () == current_query_forward_strand) {
      return false;
    } else {
      return true;
    }
  }

  /**
   *  Return the forward Strand of the subject EntryGroup from when the
   *  Comparator was created.
   **/
  public Strand getOrigSubjectForwardStrand () {
    return orig_subject_forward_strand;
  }

  /**
   *  Return the forward Strand of the query EntryGroup from when the
   *  Comparator was created.
   **/
  public Strand getOrigQueryForwardStrand () {
    return orig_query_forward_strand;
  }

  /**
   *  Arrange for the two FeatureDisplay objects to scroll in parallel.
   **/
  public void lockDisplays () {
    displays_are_locked = true;
    repaintCanvas ();
  }

  /**
   *  Arrange for the two FeatureDisplay objects to scroll independently.
   **/
  public void unlockDisplays () {
    displays_are_locked = false;
    repaintCanvas ();
  }

  /**
   *  Return true if and only if the displays are currently locked.
   **/
  public boolean displaysAreLocked () {
    return displays_are_locked;
  }

  /**
   *  Toggle whether the two displays are locked.
   **/
  public void toggleDisplayLock () {
    displays_are_locked = !displays_are_locked;
    repaintCanvas ();
  }

  /**
   *  Calling this method will temporarily disable selectFromQueryRange() and
   *  selectFromSubjectRange() until enableSelection().  This is need to allow
   *  the selections of the top and bottom FeatureDisplays to be set without
   *  changing which AlignMatches are selected.
   **/
  public void disableSelection () {
    disable_selection_from_ranges = true;
  }

  /**
   *  Enable selectFromQueryRange() and selectFromSubjectRange().
   **/
  public void enableSelection () {
    disable_selection_from_ranges = false;
  }

  /**
   *  Convert a base position into a screen x coordinate.
   **/
  private int getScreenPosition (final int canvas_width,
                                 final DisplayAdjustmentEvent event,
                                 final int base_position) {
    // this is the base that is at the left of the screen
    final int screen_start_base = event.getStart ();

    final double base_pos_float =
      event.getBaseWidth () * (base_position - screen_start_base);

    if (base_pos_float > 30000) {
      return 30000;
    } else {
      if (base_pos_float < -30000) {
        return -30000;
      } else {
        return (int) base_pos_float;
      }
    }
  }

  /**
   *  Return an array of colours that will be used for colouring the matches
   *  (depending on score).
   **/
  private Color [] makeColours () {
    final Color [] return_colours = {
      new Color (255,191,191),
      new Color (255,175,175),
      new Color (255,159,159),
      new Color (255,143,143),
      new Color (255,127,127),
      new Color (255,111,111),
      new Color (255,95,95),
      new Color (255,79,79),
      new Color (255,63,63),
      new Color (255,47,47),
      new Color (255,31,31),
      new Color (255,15,15),
      new Color (255,0,0),
    };

    return return_colours;
  }

  /**
   *  The comparison data that will be displayed in this component.
   **/
  final private ComparisonData comparison_data;

  /**
   *  All the AlignMatch objects from comparison_data (possibly in a
   *  different order.
   **/
  private AlignMatch [] all_matches = null;

  /**
   *  This is the last DisplayAdjustmentEvent reference that was passed to
   *  setSubjectSeqeuencePosition ().
   **/
  private DisplayAdjustmentEvent last_subject_event;

  /**
   *  This is the last DisplayAdjustmentEvent reference that was passed to
   *  setQuerySeqeuencePosition ().
   **/
  private DisplayAdjustmentEvent last_query_event;

  /**
   *  The FeatureDisplay that is above this component.  (From the constructor).
   **/
  private FeatureDisplay subject_feature_display;

  /**
   *  The FeatureDisplay that is below this component.  (From the constructor).
   **/
  private FeatureDisplay query_feature_display;

  /**
   *  Set by the constructor to be the original forward strand for the subject
   *  sequence.  This is use to determine whether to subject sequence has been
   *  reverse-complemented or not.
   **/
  private Strand orig_subject_forward_strand;

  /**
   *  Set by the constructor to be the original forward strand for the query
   *  sequence.  This is use to determine whether to query sequence has been
   *  reverse-complemented or not.
   **/
  private Strand orig_query_forward_strand;

  /**
   *  One of the two Entry objects that we are comparing.
   **/
  final private EntryGroup subject_entry_group;

  /**
   *  One of the two Entry objects that we are comparing.
   **/
  final private EntryGroup query_entry_group;

  /**
   *  The selected matches.  null means no matches are selected.
   **/
  private AlignMatchVector selected_matches = null;

  /**
   *  The objects that are listening for AlignmentSelectionChangeEvents.
   **/
  private java.util.Vector selection_change_listeners =
    new java.util.Vector ();

  /**
   *  The spectrum of colours use to display the scores of matches.
   **/
  private Color [] score_colours = null;

  /**
   *  The scroll bar used to set the minimum length of the visible matches.
   **/
  private Scrollbar scroll_bar = null;

  /**
   *  Matches with scores below this values will not be shown.
   **/
  private int minimum_score = 0;

  /**
   *  Matches with scores above this values will not be shown.
   **/
  private int maximum_score = 100;

  /**
   *  True if we should offer to flip the query sequence when the user
   *  double clicks on a flipped match.
   **/
  private boolean offer_to_flip_flag = false;

  /**
   *  A Vector of those objects that are listening for AlignmentEvents
   **/
  private java.util.Vector alignment_event_listeners =
    new java.util.Vector ();

  /**
   *  If true then the FeatureDisplays above and below this AlignmentViewer
   *  should scroll together.
   **/
  private boolean displays_are_locked = true;

  /**
   *  Setting this to true will temporarily disable selectFromQueryRange() and
   *  selectFromSubjectRange() until enableSelection().  This is need to allow
   *  the selections of the top and bottom FeatureDisplays to be set without
   *  changing which AlignMatches are selected.
   **/
  private boolean disable_selection_from_ranges = false;
}
